import numpy as np
import numpy.linalg as la
from sklearn.linear_model import LinearRegression


def all_close_to(a, value=0, tol=1e-7):
    return np.all(abs(a - value) < tol)


def v2a(a):
    return a.reshape((len(a), 1))


def lam_max(x, y, alpha):
    """
    Approximate the lowest value for lambda where all the coefficients will be zero

    :param x:
    :param y:
    :param alpha:
    :return:
    """
    n = len(y)
    dots = np.zeros(x.shape[1])
    for j in range(len(dots)):
        dots[j] = x[:, j].T @ y
    return np.abs(dots).max() / (n*alpha)


def soft_thres(x, thres):
    return np.sign(x) * np.maximum(np.abs(x) - thres, 0)


def quad_solution(u, v, w):
    temp = np.sqrt((v**2) - (4*u*w))
    root1 = (-v + temp) / (2*u)
    root2 = (-v - temp) / (2*u)
    return root1, root2


def solve_abg(beta_j, theta_j, grad_beta, grad_theta, alpha, lam, t):
    """
    Solves a and b so gradient iterations of theta are not needed. Equation (22)

    :param beta_j:
    :param theta_j:
    :param grad_beta:
    :param grad_theta:
    :param alpha:
    :param lam:
    :param t: Backtracking parameter. Controlled in the fit loop currently set to 0.1
    :return:
    """
    # Convergence hyperparameters
    big, eps = 10e9, 1e-3

    g1 = np.abs(beta_j - t * grad_beta)

    g2_thres = t * alpha * lam
    c = t * (1 - alpha) * lam

    scrat = soft_thres(theta_j - t * grad_theta, g2_thres)
    g2 = np.sqrt(scrat @ scrat)  # Is this fast than an l2 norm?

    root1, root2 = quad_solution(1, 2*c, 2*c*g2-g1**2-g2**2)

    a = np.array([
        g1 * root1 / (c + root1),
        g1 * root2 / (c + root2),
        g1 * root1 / (c + root2),
        g1 * root2 / (c + root1)
    ])
    b = np.array([
        root1 * (c - g2) / (c + root1),
        root2 * (c - g2) / (c + root2),
        root1 * (c - g2) / (c + root2),
        root2 * (c - g2) / (c + root1)
    ])

    x_min = big
    j_hat, k_hat = 0, 0
    for j in range(4):
        for k in range(4):
            denominator = np.sqrt(a[j]**2 + b[k]**2)
            if denominator > 0:
                val1 = (1 + (c / denominator)) * a[j] - g1
                val2 = (1 + c * (1 / b[k] + 1 / denominator)) * b[k] - g2

                temp = np.abs(val1) + np.abs(val2)
                if temp < x_min:
                    j_hat, k_hat = j, k
                    x_min = temp

    if np.abs(x_min) > eps:
        print('Failed to Solve ABG')

    if a[j_hat] < 0 or b[k_hat] < 0:
        print('Warning! One of the norms are negative [Solve ABG]')

    xnorm = np.sqrt(a[j_hat]**2 + b[k_hat]**2)

    beta_j_hat = (beta_j - t * grad_beta) / (1 + c / xnorm)

    scrat = theta_j - t * grad_theta
    theta_j_hat = soft_thres(scrat, g2_thres)
    theta_j_hat = theta_j_hat / (1 + c * (1/xnorm + 1/b[k_hat]))

    return beta_j_hat, theta_j_hat


def estimate_intercepts(z, y):
    lm = LinearRegression()
    lm.fit(z, y)

    beta_0 = lm.intercept_
    theta_0 = lm.coef_
    y = y - lm.predict(z)
    return beta_0, theta_0, y


# noinspection PyPep8Naming
class PliableLassoModelHelper:
    """
    This class allows for runtime optimised coordinate descent by trying to more aggressively cache results
    """
    def __init__(self, X=None, Z=None):
        self.w_j_list = None  # Dont use an if else statement since compute_w_j() needs this to exist first.
        if X is not None and Z is not None:
            self.w_j_list = [self.compute_w_j(X, Z, j) for j in range(X.shape[1])]

    def model(self, beta_0, theta_0, beta, theta, x, z):
        """
        The pliable lasso model described in the paper
        y ~ f(x)

        formulated as

        y ~ b_0 + Z theta_0 + X b + \sum( w_j theta_ji )

        :param beta_0:
        :param theta_0:
        :param beta:
        :param theta:
        :param x:
        :param z:
        :return:
        """
        n, p, k = x.shape[0], x.shape[1], z.shape[1]

        intercepts = beta_0 + (z @ theta_0)

        shared_model = x @ beta

        pliable = np.zeros(n)
        # For performance, check if there are even nonzero values in theta
        if np.any(theta):
            # At least 1 nonzero value in theta
            for j_i in range(p):
                # For performance, screen if theta_j is nonzero before computing pliable
                if np.any(theta[j_i, :]):
                    w_j = self.compute_w_j(x, z, j_i)
                    pliable = pliable + (w_j @ theta[j_i, :])

        return intercepts + shared_model + pliable

    def partial_model(self, beta_0, theta_0, beta, theta, x, z, ignore_j):
        """
        y ~ f(x) with X_j removed from the model

        :param beta_0:
        :param theta_0:
        :param beta:
        :param theta:
        :param x:
        :param z:
        :param ignore_j:
        :return:
        """
        beta[ignore_j] = 0.0
        theta[ignore_j, :] = 0.0
        return self.model(beta_0, theta_0, beta, theta, x, z)

    def objective(self, beta_0, theta_0, beta, theta, x, z, y, alpha, lam):
        """
        Full objective function J(beta, theta) described in the paper

        :param beta_0:
        :param theta_0:
        :param beta:
        :param theta:
        :param x:
        :param z:
        :param y:
        :param alpha:
        :param lam:
        :return:
        """
        n = len(y)

        mse = (1 / (2 * n)) * la.norm(y - self.model(beta_0, theta_0, beta, theta, x, z), 2) ** 2
        beta_matrix = v2a(beta)

        penalty_1 = la.norm(np.hstack([beta_matrix, theta]), 2, axis=1).sum()
        penalty_2 = la.norm(theta, 2, axis=1).sum()
        penalty_3 = np.abs(theta).sum()

        cost = mse + (1 - alpha) * lam * (penalty_1 + penalty_2) + alpha * lam * penalty_3
        return cost

    def partial_objective(self, params_j, j_i, beta_0, theta_0, beta, theta, x, z, y, alpha, lam):
        """
        Computes the cost function J with feature _j removed from the model

        :param params_j:
        :param j_i:
        :param beta_0:
        :param theta_0:
        :param beta:
        :param theta:
        :param x:
        :param z:
        :param y:
        :param alpha:
        :param lam:
        :return:
        """
        beta_j, theta_j = params_j[0], params_j[1:]
        beta[j_i] = beta_j
        theta[j_i, :] = theta_j
        return self.objective(beta_0, theta_0, beta, theta, x, z, y, alpha, lam)

    def compute_w_j(self, x, z, j: int):
        """
        Performs the elementwise multiplication of X_j with Z

        :param x:
        :param z:
        :param j:
        :return:
        """
        if self.w_j_list is None:
            w_j = np.zeros(z.shape)
            for k_i in range(z.shape[1]):
                w_j[:, k_i] = x[:, j] * z[:, k_i]
            return w_j
        else:
            return self.w_j_list[j]

    def derivative_wrt_beta_j(self, beta_0, theta_0, beta, theta, x, z, y, j, alpha, lam):
        """
        Full Derivative with respect to beta_j as described in the paper

        :param beta_0:
        :param theta_0:
        :param beta:
        :param theta:
        :param x:
        :param z:
        :param y:
        :param j:
        :param alpha:
        :param lam:
        :return:
        """
        r = y - self.model(beta_0, theta_0, beta, theta, x, z)
        beta_j_theta_j = np.hstack((beta[j], theta[j, :]))
        u = 0 if all_close_to(beta_j_theta_j, 0) else beta[j] / la.norm(beta_j_theta_j, 2)
        return (-1 / len(r)) * x[:, j].T @ r + (1 - alpha) * lam * u

    def derivative_wrt_theta_j(self, beta_0, theta_0, beta, theta, x, z, y, j, alpha, lam):
        """
        Full Derivative with respect to theta_j as described in the paper

        :param beta_0:
        :param theta_0:
        :param beta:
        :param theta:
        :param x:
        :param z:
        :param y:
        :param j:
        :param alpha:
        :param lam:
        :return:
        """
        beta_j = beta[j]
        theta_j = theta[j, :]

        r = y - self.model(beta_0, theta_0, beta, theta, x, z)
        w_j = self.compute_w_j(x, z, j)

        beta_j_theta_j = np.hstack((beta_j, theta_j))
        u2 = 0 if all_close_to(beta_j_theta_j, 0) else theta_j / la.norm(beta_j_theta_j, 2)

        u3 = 0 if all_close_to(theta_j, 0) else theta_j / la.norm(theta_j, 2)

        v = np.sign(theta[j, :])

        return (-1 / len(r)) * w_j.T @ r + (1 - alpha) * lam * (u2 + u3) + alpha * lam * v



import warnings
import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.exceptions import NotFittedError

from .helpers import *


OPTIMISE_CONVEX = 'convex'
OPTIMISE_COORDINATE = 'coordinate'


# noinspection PyPep8Naming
class PliableLasso(BaseEstimator):
    """
    Pliable Lasso https://arxiv.org/pdf/1712.00484.pdf
    """
    def __init__(self, alpha=0.5, n_lam=50, max_iter=100, min_lam=0, fit_intercepts=False, verbose=False):
        self.min_lam, self.alpha = min_lam, alpha
        self.n_lam = n_lam
        self.max_iter = max_iter
        self.fit_intercepts = fit_intercepts

        # Model coefs
        self.beta_0 = None
        self.theta_0 = None
        self.beta = None
        self.theta = None

        # Metrics
        self.verbose = verbose
        self.history = []
        self.paths = {}

        # Prediction helpers
        self.static_func = PliableLassoModelHelper()

    def fit(self, X, Z, y, optimizer=OPTIMISE_COORDINATE):
        if optimizer == OPTIMISE_CONVEX:
            return self._fit_convex_optimization(X, Z, y)
        elif optimizer == OPTIMISE_COORDINATE:
            return self._fit_coordinate_descent(X, Z, y)
        else:
            raise ValueError(f'Only allowed optimizers are {OPTIMISE_COORDINATE} or {OPTIMISE_CONVEX}')

    def _fit_convex_optimization(self, X, Z, y):
        try:
            import cvxpy as cvx
            from .cvxHelpers import objective_cvx
        except ModuleNotFoundError:
            return self._fallback_to_coordinate_descent(X, Z, y)

        # Check CVXPY version number
        from distutils.version import LooseVersion
        if LooseVersion(cvx.__version__) < LooseVersion('1.0.11'):
            return self._fallback_to_coordinate_descent(X, Z, y)

        self.history = []

        # Hyperparameters
        alpha = self.alpha
        lam = cvx.Parameter(nonneg=True)
        lam.value = self.min_lam

        # Parameters
        n, p = X.shape
        k = Z.shape[1]
        if self.fit_intercepts:
            beta_0 = cvx.Variable(1)
            theta_0 = cvx.Variable(k)
        else:
            beta_0 = 0.0
            theta_0 = np.zeros(k)
        beta = cvx.Variable(p)
        theta = cvx.Variable((p, k))

        # Fit with Convex Optimisation
        problem = cvx.Problem(
            cvx.Minimize(objective_cvx(beta_0, theta_0, beta, theta, X, y, Z, alpha, lam))
        )
        # Solve on a decreasing lambda path
        problem.solve(verbose=self.verbose, solver=cvx.CVXOPT, max_iter=self.max_iter)

        self.beta_0 = beta_0 if self.fit_intercepts else beta_0.value
        self.theta_0 = theta_0 if self.fit_intercepts else theta_0.value
        self.beta = beta.value
        self.theta = theta.value

        return self

    def _fit_coordinate_descent(self, X, Z, y):
        self._reset_paths_dict()
        n, p = X.shape
        k = Z.shape[1]

        # Step 1: Setup
        alpha = self.alpha  # So I don't have to keep writing self
        beta_0, theta_0, beta, theta = 0.0, np.zeros(k), np.zeros(p), np.zeros((p, k))

        # Same lambda path specs as GLMnet (Apparently faster than solving for a single lambda
        lambda_max = lam_max(X, y, alpha)
        lambda_min = 1e-3 * lambda_max
        lambda_path = np.logspace(np.log10(lambda_max), np.log10(lambda_min), self.n_lam)
        lambda_path = lambda_path[lambda_path >= self.min_lam]
        if len(lambda_path) == 0:
            raise ValueError(f'`min_lam` was set too high! Maximum lambda for this problem is {lambda_max}')
        self.paths['lam'] = lambda_path
        self.min_lam = max(lambda_path.min(), self.min_lam)

        func = PliableLassoModelHelper(X, Z)

        # Step 2: Update coefficients
        for lam in lambda_path:  # NOTE: This means you are ignoring the self.lam value
            for i in range(self.max_iter):
                # TODO 1/14/2019 estimate beta_0 and theta_0
                beta_new, theta_new = beta.copy(), theta.copy()
                iter_prev_score = func.objective(beta_0, theta_0, beta_new, theta_new, X, Z, y, alpha, lam)

                # Iterate through all p features
                tolerance = 1e-3
                for j in range(p):
                    x_j = X[:, j]
                    r_min_j = y - func.partial_model(beta_0, theta_0, beta_new, theta_new, X, Z, j)
                    w_j = func.compute_w_j(X, Z, j)

                    cond_17a = np.abs(x_j.T @ r_min_j / n) <= (1-alpha) * lam
                    cond_17b = la.norm(soft_thres(w_j.T @ r_min_j / n, alpha * lam), 2) <= 2 * (1-alpha) * lam

                    if cond_17a and cond_17b:
                        if self.verbose >= 4:
                            print(f'beta_{j} == 0 and theta_{j} == 0')

                    else:
                        beta_j_hat = (n / la.norm(x_j, 2)**2) * soft_thres(x_j.T @ r_min_j / n, (1-alpha) * lam)

                        cond_19 = la.norm(soft_thres(w_j.T @ (r_min_j - x_j * beta_j_hat) / n, alpha * lam), 2)
                        cond_19 = cond_19 <= (1-alpha) * lam

                        if cond_19:
                            beta_new[j] = beta_j_hat
                            if self.verbose >= 2:
                                print(f'beta_{j} != 0 and theta_{j} == 0')
                                print(f'-> {beta_j_hat}')

                        else:
                            if self.verbose >= 2:
                                print(f'beta_{j} != 0 and theta_{j} != 0')
                            t, l, eps = 0.1, 1.0, 1e-5
                            max_steps = 100
                            objective_prev = func.objective(beta_0, theta_0, beta_new, theta_new, X, Z, y, alpha, lam)
                            for _ in range(max_steps):
                                r = y - func.model(beta_0, theta_0, beta_new, theta_new, X, Z)
                                beta_j_hat = beta_new[j]  # Was this a problem?
                                theta_j_hat = theta_new[j, :]

                                grad_beta_j = -np.sum(x_j * r) / n
                                grad_theta_j = -w_j.T @ r / n
                                if self.verbose >= 3:
                                    print('--> Gradients')
                                    print(grad_beta_j)
                                    print(grad_theta_j)

                                beta_j_new_hat, theta_j_new_hat = solve_abg(
                                    beta_j_hat, theta_j_hat,
                                    grad_beta_j, grad_theta_j,
                                    alpha, lam, t
                                )

                                if self.verbose >= 3:
                                    print(f'new beta_{j} =', beta_j_new_hat)
                                    print(f'new theta_{j} =', theta_j_new_hat)

                                # Update Coefs
                                beta_new[j] = beta_j_new_hat
                                theta_new[j, :] = theta_j_new_hat
                                objective_current = func.objective(
                                    beta_0, theta_0, beta_new, theta_new, X, Z, y, alpha, lam
                                )
                                improvement = objective_prev - objective_current

                                if abs(improvement) < tolerance:
                                    if self.verbose >= 2:
                                        print(f'==> Step converged j = {j} : {improvement}')
                                    break
                                else:
                                    objective_prev = objective_current

                            else:
                                warnings.warn(f'==> Max Steps reached (Failed to converged) on beta_{j}')

                    beta, theta = beta_new.copy(), theta_new.copy()

                if self.verbose >= 1:
                    from sklearn.metrics import r2_score
                    score_i = r2_score(y, func.model(beta_0, theta_0, beta, theta, X, Z))
                    print(f'==> Iter {i} @ lam {lam:0.3f} = {score_i:0.2%}')

                iter_current_score = func.objective(beta_0, theta_0, beta_new, theta_new, X, Z, y, alpha, lam)
                if abs(iter_prev_score - iter_current_score) < tolerance:
                    if self.verbose >= 1:
                        print(f'==> Converged on lam_i = {lam:0.3f}\n')
                    break
                else:
                    if self.verbose >= 1:
                        print(f'delta J_{i} = {iter_prev_score - iter_current_score}')
                    iter_prev_score = iter_current_score

            # End of lam_i iteration. Update Paths
            self.paths['beta_0'].append(beta_0)
            self.paths['theta_0'].append(theta_0)
            self.paths['beta'].append(beta)
            self.paths['theta'].append(theta)

        # Finally update model's parameters
        self.beta_0, self.theta_0, self.beta, self.theta = beta_0, theta_0, beta, theta
        for var in ['theta_0', 'beta', 'theta']:
            self.paths[var] = np.stack(self.paths[var])

    def _fallback_to_coordinate_descent(self, X, Z, y):
        warnings.warn('cvxpy is required for convex optimisation. Falling back to coordinate descent')
        return self._fit_coordinate_descent(X, Z, y)

    def _reset_paths_dict(self):
        self.paths = {}
        for var in ['lam', 'beta_0', 'theta_0', 'beta', 'theta']:
            self.paths[var] = []

    def predict(self, X, Z):
        if self.beta is None:
            raise NotFittedError

        if isinstance(X, pd.Series) or isinstance(X, pd.DataFrame):
            X = X.values

        if isinstance(Z, pd.Series) or isinstance(Z, pd.DataFrame):
            Z = Z.values

        return self.static_func.model(self.beta_0, self.theta_0, self.beta, self.theta, X, Z)

    def score(self, X, Z, y):
        from sklearn.metrics import r2_score
        return r2_score(y, self.static_func.model(self.beta_0, self.theta_0, self.beta, self.theta, X, Z))

    def cost(self, X, Z, y):
        return self.static_func.objective(
            self.beta_0, self.theta_0, self.beta, self.theta, X, Z, y, self.alpha, self.min_lam
        )



import numpy as np
import cvxpy as cvx
from .helpers import v2a


def compute_w_j(x, z, j):
    k = z.shape[1]
    x_j = np.tile(v2a(x[:, j]), (1, k))
    return x_j * z


def y_hat(beta_0, theta_0, beta, theta, x, z):
    # Confirmed DCP and Convex
    n, p = x.shape

    intercepts = beta_0 + (z @ theta_0)

    shared_model = x @ beta

    pliables = np.zeros(n)
    for j_i in range(p):
        w_j = compute_w_j(x, z, j_i)
        pliables = pliables + (w_j @ theta[j_i, :])

    output = intercepts + shared_model + pliables
    return output


def objective_cvx(beta_0, theta_0, beta, theta, x, y, z, alpha, lam):
    n, p = x.shape

    # mse = (1/(2*n)) * cvx.sum((y - y_hat(beta_0, theta_0, beta, theta, x, z)) ** 2)
    mse = cvx.norm((y - y_hat(beta_0, theta_0, beta, theta, x, z)), p=2)**2

    beta_matrix = cvx.reshape(beta, (p, 1))

    penalty_1 = cvx.sum(cvx.norm(cvx.hstack([beta_matrix, theta]), p=2, axis=1))
    penalty_2 = cvx.sum(cvx.norm(theta, p=2, axis=1))
    penalty_3 = cvx.sum(cvx.norm(theta, p=1))

    loss = mse + (1 - alpha) * lam * (penalty_1 + penalty_2) + alpha * lam * penalty_3
    return loss